#!/usr/bin/env python


#############################################################################
#                                                                           #
#   Vital_Signs.py                                                          #
#                                                                           #
#   Uses a Sense HAT's LEDs to display columns of:                          #
#       Either:                                                             #
#           The combined load % across all CPU cores, or                    #
#           Each core's load %.                                             #
#       Memory available/unavailable                                        #
#       CPU temperature (C or F)                                            #
#       Sense HAT's temperature (C or F)                                    #
#       Sense HAT's % humidity                                              #
#                                                                           #
#   Optionally displays ping results as OK/fail for up to 8 computers.      #
#                                                                           #
#   The data this app outputs is organized by column. The Sense HAT's LED   #
#       array is organized by row. The least-complicated programming        #
#       methodology is to map columns to the LED array's rows. The first    #
#       row at the top becomes the left column.                             #
#                                                                           #
#   The terms "LED" and "pixel" are used interchangeably. In most cases     #
#       LED is used when referencing hardware and pixel is used for arrays. #
#                                                                           #
#   The columns appear 90 degrees clockwise from the rows. To see the       #
#       columns in the row's orientation the columns' LED array is rotated  #
#       -90 (270) degrees.                                                  #
#                                                                           #
#   I went overkill on features, the interface is way to busy.  I did this  #
#       to hone my programming skills in Python, Raspbian, the Python IDE,  #
#       and the Sense HAT.                                                  #
#                                                                           #
#   The Ping Hosts feature runs as an independent thread on its own update  #
#       schedule. That way if it gets delayed waiting for responses the     #
#       main display will update on schedule. The main thread displays the  #
#       ping results regardless of how far through its cycle the Ping Hosts #
#       thread is.                                                          #
#                                                                           #
#   The Apple logo is displayed because "I'm a Mac". My Raspberry Pi's run  #
#       headless (without a display, mouse, or keyboard). I control them    #
#       using VNC Viewer on my Mac.                                         #
#                                                                           #
#   Variables et al. have their first letter capitalized and use            #
#       underscores between words. This makes it clearer what I defined vs  #
#       what's in Python, various modules, and the Sense HAT's API.         #
#                                                                           #
#   Stop execution by typing ctrl-c.                                        #
#                                                                           #
#   Version History                                                         #
#       1.0.0 2019-03-30 - Initial release.                                 #
#       1.0.1 2019-04-03 - Cleaned up comments, renamed some variables.     #
#       1.0.2 2019-04-26 - Added the Ping Hosts feature.                    #
#       1.0.3 2019-04-26 - Bug fixes.                                       #
#                                                                           #
#   Samuel Herschbein                                                       #
#   sam@SamIsIT.com                                                         #
#                                                                           #
#############################################################################

#############################################################################
#                                                                           #
#   BEGIN - parameters not requiring reprogramming                          #
#                                                                           #
#############################################################################

                        #   Dim the display if your RPi is in a darker environment.
Low_Light_Mode_On = True

                        #   My RPi's have their SD card side facing me.
                        #   No error checking, use 0, 90, 180, or 270.
LED_Array_Rotation = 0

                        #   How long to display each ring in the TV Effect.
Seconds_Between_TV_On_Off_Effect_Steps = 0.15

                        #   How long to display the Raspberry Pi logo at launch and exit.
Seconds_to_Display_RPi_Logo = 3

                        #   How long to display the Apple logo at launch and exit.
Seconds_to_Display_Apple_Logo = 3


                        #   These flags let you decide which stats you want displayed and in which columns.
                        #   Column numbers are 0..7. The CPU_Load occupies 4 columns.
                        #   There is no error checking.
CPU_Load_Display_On = True
CPU_Load_Display_Column_1 = 0      #   Last column = CPU_Load_Display_Column_1 + 3.

Memory_Display_On = True
Memory_Display_Column = 4

CPU_Temp_Display_On = True
CPU_Temp_Display_Column = 5

HAT_Temp_Display_On = True
HAT_Temp_Display_Column = 6

Humidity_Display_On = True
Humidity_Display_Column = 7


                        #   When True each core's load % will be displayed.
                        #   When False, the combined load % for all cores will be displayed in a single column
                        #       which will scroll left to show history (granted, only 4 points).
CPU_by_Core = True
                        #   When CPU_by_Core = True and the RPi has only one core: if CPU_by_Core_Single_Core_2_Columns = True
                        #       the load will be displayed in two identical columns.
                        #   Otherwise the load will be displayed in the usual single column.
CPU_by_Core_Single_Core_2_Columns =  True

                        #   Personal preference. Since this is a float you can use fractions of a second.
Seconds_Between_Updates = 1.75

                        #   The "It's Alive" feature flashes a row using colors defined below. The row ping-pongs between the top to bottom.
                        #   Since this overwrites the display the flash should be brief to minimize visual confusion.
                        #   This shows that the app is running when the displayed values aren't changing.
                        #   Read the "Ping Hosts" section below, it is an addition to It's Alive.
Its_Alive_On = True
                        #   When Ping_Hosts_Display_On = True this should be higher to improve readability.
                        #   I prefer 0.5 when True and 0.25 when False.
Its_Alive_Seconds_to_Flash = 0.5

                        #   This feature only works when Its_Alive_On = True.
                        #   Each host in the list below is pinged to see if it's alive or not.
                        #   This replaces the colors used by It's Alive with either an OK or failed color.
Ping_Hosts_Display_On = True
                        #   The Linux ping command can be specified with a wait time. Only a single packet is requested since
                        #       the response is parsed with grep and awk.
                        #   For local subnets, 2 seconds is a resonable wait time. When pinging hosts on external networks
                        #       that are several hops away this may need to be increased. The downside to increasing it
                        #       too much is longer wait times when hosts are unreachable.
                        #   The ping command didn't work correctly on my RPi when the wait time was set to 0.
                        #   This is specified as a string because it's inserted into the ping command's string.
Ping_Hosts_Wait_Time = "2"
                        #   Ping Hosts runs as a separate thread on its own update schedule.
                        #   For most uses polling every 10 seconds is adequate.
                        #   I don't recommend setting this to 0, the thread will constantly loop and could use too
                        #       many resources.
Seconds_Between_Pings = 10
                        #   List of 8 host IPs/Domains to ping to see if they're alive.
                        #   The list must have 8 entries, use "" for unused entries.
                        #   Unfilled entries have their pixel set using the It's Alive color for that pixel.
                        #   It doesn't matter which entries are blank and which aren't. Be creative!
Ping_Hosts_IP_List = [
    "10.20.40.1",       #   Host 1 -
    "10.20.40.2",       #   Host 2 -
    "10.20.40.3",       #   Host 3 -
    "10.20.40.4",       #   Host 4 -
    "10.20.40.5",       #   Host 5 -
    "10.20.40.6",       #   Host 6 -
    "10.20.40.7",       #   Host 7 -
    "10.20.40.8"        #   Host 8 -
    ]

                        #   The CPU loads and temperatures are displayed using a rainbow spectrum which
                        #       varies from 0..8 where 0 means the pixel's LED is off (black).
                        #   This option allows a color outside of the spectrum so near 0 values are visible.
                        #   A shade of grey is the best choice since grays aren't part of the color spectrum.
Separate_Color_For_0 = True

                        #   Uses Fahrenheit when True, Celsius when False.
CPU_Temp_Fahrenheit = True
                        #   The cutoffs must be in the same units as the temperature.
                        #   The values are float, so 123.45 is valid.
                        #   The cutoffs are used as a range for the LEDs.
                        #       The range is 0..8 with the LED off being 0.
                        #       Temperatures above the high cutoff map to 8.
                        #       Temperatures below the low cutoff map to 0.
CPU_Temp_Cutoff_High = 150
CPU_Temp_Cutoff_Low = 90

                        #   Uses Fahrenheit when True, Celsius when False.
HAT_Temp_Fahrenheit = True
                        #   The cutoffs must be in the same units as the temperature.
                        #   The values are float, so 123.45 is valid.
                        #   The cutoffs are used as a range for the LEDs.
                        #       The range is 0..8 with the LED off being 0.
                        #       Temperatures above the high cutoff map to 8.
                        #       Temperatures below the low cutoff map to 0.
HAT_Temp_Cutoff_High = 120
HAT_Temp_Cutoff_Low = 50



#############################################################################
#                                                                           #
#   END - parameters not requiring reprogramming                            #
#                                                                           #
#############################################################################


#############################################################################
#                                                                           #
#   BEGIN - Define colors                                                   #
#                                                                           #
#############################################################################
                                #   The Sense HAT's LEDs don't use the full 0..255 range.
                                #       From what I've read only the upper 5 bits are sent to the physical LED hardware.
                                #       The largest # that matters is 248 (1111 1000, 0xF8).
                                #       This means the brightest near white should be (248, 248, 248) and the darkest near grey
                                #       should be (8, 8, 8). Reality is different.
                                #
                                #   When Low Light is on 8 is the dimmest single color value visible.
                                #   When Low Light is off 48 is the dimmest single color value visible. This isn't obvious
                                #       since 48 (0011 0000, 0x30) has 2 bits set.
                                #   I haven't dissected the driver code to figure this out.

                                #   I tried having separate LED colors for both Low Light on and off, it wasn't worth the effort
                                #       of calculating the correct values. The Sense API does a good enough job of mapping colors
                                #       to Low Light levels.
                                #   If you want to have separate colors for Low Light it's easy enough, here's an example:
                                #       my_dual_mode_color = [(R_High, G_High, B_High), (R_Low, G_Low, B_Low)][Low_Light_Mode_On]
                                #       The boolean works for array indexing since in Python False == 0 and True == 1.
                                #       Thus the first tuple is for when Low Light is off, the second when it's on.

                                #   Since bits are stripped off I'll specify RGBs using hex notation with 2 digits. This made it
                                #       easier for me to visualize the bits.

Black = (0x00, 0x00, 0x00)

                                #   Shades of grey nearly even spaced between brightest and dimmest.
                                #   Grey_Level_0 is the dimmest gray my Sense HAT will display.
Grey_Level_7 = (0xF8, 0xF8, 0xF8)
Grey_Level_6 = (0xD8, 0xD8, 0xD8)
Grey_Level_5 = (0xC0, 0xC0, 0xC0)
Grey_Level_4 = (0xA0, 0xA0, 0xA0)
Grey_Level_3 = (0x80, 0x80, 0x80)
Grey_Level_2 = (0x68, 0x68, 0x68)
Grey_Level_1 = (0x48, 0x48, 0x48)
Grey_Level_0 = (0x30, 0x30, 0x30)

                                #   These RGB values for the loads & temperatures use a rainbow spectrum.
                                #   Technically violet is hotter than red, but we humans think of reds as hotter than blues.
Spectrum_Level_7 = (0xA0, 0x00, 0x00)   #   Red
Spectrum_Level_6 = (0xC8, 0x60, 0x00)   #   Orange
Spectrum_Level_5 = (0xC8, 0xC8, 0x00)   #   Yellow
Spectrum_Level_4 = (0x68, 0xC8, 0x00)   #   Light Green
Spectrum_Level_3 = (0x00, 0x88, 0x00)   #   Dark Green
Spectrum_Level_2 = (0x00, 0x60, 0xC0)   #   Blue
Spectrum_Level_1 = (0x00, 0x00, 0xA8)   #   Indigo
Spectrum_Level_0 = (0x40, 0x00, 0x68)   #   Violet

                                #   These RGB values are for the humidity.
                                #   I chose shades of blue since blue is associated with water.
Humidity_Level_7 = (0x00, 0x00, 0xF8)
Humidity_Level_6 = (0x00, 0x00, 0xD8)
Humidity_Level_5 = (0x00, 0x00, 0xC0)
Humidity_Level_4 = (0x00, 0x00, 0xA0)
Humidity_Level_3 = (0x00, 0x00, 0x80)
Humidity_Level_2 = (0x00, 0x00, 0x68)
Humidity_Level_1 = (0x00, 0x00, 0x48)
Humidity_Level_0 = (0x00, 0x00, 0x30)

                                #   These RGB values are for the memory available and unavailable.
Mem_Avail_Color = (0x00, 0x4C, 0x4C)        #   Dark cyan
Mem_Unavailable_Color = (0x70, 0x38, 0x00)  #   Dark orange

                                #   It's Alive LED row. You can change what appears in each LED.
                                #   I'd avoid black since black is a common color. I recommend shades of grey.
                                #   The leftmost pixel is first.
                                #   This has to be specified using [] around each RGB tuple to work with concatenation of tuples.
Its_Alive_Row = [
    [Grey_Level_2],
    [Grey_Level_2],
    [Grey_Level_2],
    [Grey_Level_4],
    [Grey_Level_4],
    [Grey_Level_2],
    [Grey_Level_2],
    [Grey_Level_2]
    ]

                                #   These RGB values are for successful & failed pinging of remote hosts.
                                #   This is defined as a list since it's indexed using the result of the ping, either 0 or 1.
                                #   This has to be specified using [] around each RGB tuple to work with concatenation of tuples.
Ping_Hosts_Colors_List = [
    [(0xF8, 0x00, 0x00)],       #   0 - Ping failed - bright red.
    [(0x00, 0x88, 0x00)]        #   1 - Ping succeeded - dark green.
    ]

                                #   The color to use to indicate that an LED is not lit.
                                #   A shade of grey is the best choice since grays aren't part of the color spectrum.
Color_For_0 = Grey_Level_0

                                #   Grey values to use for the TV visual effect.
                                #   Listed here from dim to bright.
                                #   The routine will reverse the order when turning off.
TV_On_Off_Effect_Grey_Levels = [
    Grey_Level_1,
    Grey_Level_3,
    Grey_Level_5,
    Grey_Level_7
    ]

#############################################################################
#                                                                           #
#   END - Define colors.                                                    #
#                                                                           #
#############################################################################


#############################################################################
#                                                                           #
#   BEGIN - Program code.                                                   #
#                                                                           #
#############################################################################

import psutil                   #   Python module for getting CPU and memory stats from the OS.

import subprocess               #   Python module that allows calling command line ping.

import threading                #   Pythin module for multi-threading. Used to setup asynchronous Ping daemon.

from time import sleep          #   Python module with time functions. Sleep is used to delay between operations.

from sense_hat import SenseHat  #   Sense HAT API.
sense = SenseHat()              #   Everyone seems to use "sense" so it's common sense to use the common "sense."


#   Populate the RPi logo LED array.
#   The RGBs from the official Raspberry Pi logo are:
#       RPi_Logo_Red = (196, 25, 73)
#       RPi_Logo_Green = (107, 192, 72)
#   To my eyes they don't look right on the Sense HAT's LEDs so I tweaked them:
#       RPi_Logo_Red = (128, 0, 48)
#       RPi_Logo_Green = (64, 160, 64)
RPi_Logo_Red = (0x80, 0x00, 0x30)
RPi_Logo_Green = (0x40, 0xB0, 0x40)
RPi_Logo_LED_Array = [
    RPi_Logo_Green, RPi_Logo_Green, Black,          Black,          Black,          Black,          RPi_Logo_Green, RPi_Logo_Green,
    Black,          RPi_Logo_Green, RPi_Logo_Green, Black,          Black,          RPi_Logo_Green, RPi_Logo_Green, Black,
    Black,          RPi_Logo_Red,   Black,          RPi_Logo_Red,   RPi_Logo_Red,   Black,          RPi_Logo_Red,   Black,
    RPi_Logo_Red,   Black,          RPi_Logo_Red,   Black,          Black,          RPi_Logo_Red,   Black,          RPi_Logo_Red,
    RPi_Logo_Red,   Black,          RPi_Logo_Red,   Black,          Black,          RPi_Logo_Red,   Black,          RPi_Logo_Red,
    Black,          RPi_Logo_Red,   Black,          RPi_Logo_Red,   RPi_Logo_Red,   Black,          RPi_Logo_Red,   Black,
    Black,          Black,          RPi_Logo_Red,   Black,          Black,          RPi_Logo_Red,   Black,          Black,
    Black,          Black,          Black,          RPi_Logo_Red,   RPi_Logo_Red,   Black,          Black,          Black
    ]

#   Populate the Apple logo LED array.
#   I started with a 16x16 logo taken from Mac OS X 10.6 Snow Leopard's Core Types.
#   In Photoshop I added a Black background and then converted it to 8x8 using Nearest Neighbor resampling.
#   A few pixels were far off and were manually edited.
#   I exported it as a PNG and uploaded it to the RPi.
#   In another Python script I used sense.load_image to load it on the LEDs.
#   Then I used sense.get_pixels to get it in an array.
#   At the Python command prompt I entered the array name.
#   I then copied the result into this app and edited to get the array format.
#   Finally I tweaked a few pixels.
Apple_Logo_LED_Array = [
    (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x10, 0x78, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00),
    (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x10, 0x78, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00),
    (0x00, 0x00, 0x00), (0x10, 0x78, 0x00), (0x10, 0xB0, 0x00), (0x10, 0xB0, 0x00), (0x10, 0xB0, 0x00), (0x10, 0x78, 0x00), (0x00, 0x48, 0x00), (0x00, 0x00, 0x00),
    (0xD0, 0xA0, 0x20), (0x00, 0x00, 0x00), (0xF8, 0xC0, 0x28), (0xF8, 0xC0, 0x28), (0xF8, 0xC0, 0x28), (0x80, 0x50, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00),
    (0xB8, 0x50, 0x00), (0x00, 0x00, 0x00), (0xF8, 0x84, 0x08), (0xB8, 0x50, 0x00), (0xB8, 0x50, 0x00), (0x80, 0x20, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00),
    (0xD0, 0x00, 0x00), (0x00, 0x00, 0x00), (0xD0, 0x00, 0x00), (0xD0, 0x00, 0x00), (0xD0, 0x00, 0x00), (0xA8, 0x00, 0x00), (0x50, 0x00, 0x00), (0x00, 0x00, 0x00),
    (0x00, 0x00, 0x00), (0xB8, 0x18, 0x80), (0xB8, 0x18, 0x80), (0xB8, 0x18, 0x80), (0xB8, 0x18, 0x80), (0x80, 0x18, 0x80), (0x40, 0x00, 0x40), (0x00, 0x00, 0x00),
    (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x90), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00), (0x00, 0x00, 0x90), (0x00, 0x00, 0x00), (0x00, 0x00, 0x00)
    ]

#   Fill a column with the color spectrum.
Color_Spectrum_Column = [Spectrum_Level_0, Spectrum_Level_1, Spectrum_Level_2, Spectrum_Level_3, Spectrum_Level_4, Spectrum_Level_5, Spectrum_Level_6, Spectrum_Level_7]

#   Fill a column with the humidity colors.
Humidity_LED_Column = [Humidity_Level_0, Humidity_Level_1, Humidity_Level_2, Humidity_Level_3, Humidity_Level_4, Humidity_Level_5, Humidity_Level_6, Humidity_Level_7]

#   Fill a Black column for unused colors.
Black_Column = [Black, Black, Black, Black, Black, Black, Black, Black]

#   Fill an all Black LED array.
Black_LED_Array = Black_Column + Black_Column + Black_Column + Black_Column + Black_Column + Black_Column + Black_Column + Black_Column

#   Fill the column used when 0's black is replaced by a color.
Color_For_0_Column = [Color_For_0, Black, Black, Black, Black, Black, Black, Black]

#   Fill columns with the two memory colors.
Mem_Avail_Column = [Mem_Avail_Color, Mem_Avail_Color, Mem_Avail_Color, Mem_Avail_Color, Mem_Avail_Color, Mem_Avail_Color, Mem_Avail_Color, Mem_Avail_Color]
Mem_Unavailable_Column = [Mem_Unavailable_Color, Mem_Unavailable_Color, Mem_Unavailable_Color, Mem_Unavailable_Color, Mem_Unavailable_Color, Mem_Unavailable_Color, Mem_Unavailable_Color, Mem_Unavailable_Color]

#   Setup the column to use when all LEDs are off.
Column_For_0_Value =[Black_Column, Color_For_0_Column][Separate_Color_For_0]

#   Define functions used repeatedly.

#   Normalize an integer so it's between 0 and 8.
def Normalize_to_0_Thru_8 (The_Integer):
    if The_Integer < 0:
        The_Integer = 0
    elif The_Integer > 8:
        The_Integer = 8
    return (The_Integer)

#   Visual effect mimicking how old B&W TVs looked when they were turned on & off.
def TV_On_Off_Effect (Turn_TV_On):
    if Turn_TV_On:
        sense.set_pixels(Black_LED_Array)                   #   Start with black array.
        for This_Step in range (0, 4):                      #   4 steps, add from inner ring outwards.
            for H_Index in range (3 - This_Step, 5 + This_Step):
                for V_Index in range (3 - This_Step, 5 + This_Step):
                    sense.set_pixel(H_Index, V_Index, TV_On_Off_Effect_Grey_Levels[This_Step])
            sleep(Seconds_Between_TV_On_Off_Effect_Steps)
    else:
        for This_Step in range (3, -1, -1):                 #   4 steps, remove from outer ring inwards.
            sense.set_pixels(Black_LED_Array)               #   Easiest to start with black and add back in.
            for H_Index in range (3 - This_Step, 5 + This_Step):
                for V_Index in range (3 - This_Step, 5 + This_Step):
                    sense.set_pixel(H_Index, V_Index, TV_On_Off_Effect_Grey_Levels[This_Step])
            sleep(Seconds_Between_TV_On_Off_Effect_Steps)
    sleep(2 * Seconds_Between_TV_On_Off_Effect_Steps)           #   Sleep a tad longer.
    sense.clear()

#   Daemon thread for pinging hosts
def Ping_Hosts_Daemon():
                                                                #   Ping Hosts runs asynchronously to the main process. It populates a row
                                                                #       which is copied by the main routine.
    global Ping_Hosts_Row                                       #   The row must be declared as global so the main thread can access it.
    while True:
        for This_Ping_Hosts_Pixel in range (0, 8):              #   Populate each column's pixel through the row.
            if Ping_Hosts_IP_List[This_Ping_Hosts_Pixel] != "":
                This_Ping_Hosts_Pixel_Color = Ping_Hosts_Colors_List[int(subprocess.getoutput("ping " + Ping_Hosts_IP_List[This_Ping_Hosts_Pixel] + \
                                     " -c 1 -W " + Ping_Hosts_Wait_Time + " -q  | grep 'received' | awk -F',' '{ print $2}' | awk '{ print $1}'"))]
            else:                                               #   If no host entered, use It's Alive's color for this pixel.
                This_Ping_Hosts_Pixel_Color = Its_Alive_Row[This_Ping_Hosts_Pixel]
            Ping_Hosts_Row = Ping_Hosts_Row[0:This_Ping_Hosts_Pixel] + [This_Ping_Hosts_Pixel_Color] + Ping_Hosts_Row[This_Ping_Hosts_Pixel + 1:8]
        sleep(Seconds_Between_Pings)


#   Main code below here.
if __name__ == '__main__':

    #   The code is in a try so interrupts (ctrl-c) can be trapped.
    #   This allows the LEDs to be turned off as the code exits.

    try:

        #   Clear anything on the Sense HAT's display
        sense.clear()

        #   Turn Low Light off while logos are displayed.
        sense.low_light = False

        #   Set the rotation based on the user's preference.
        sense.set_rotation(LED_Array_Rotation)

        #   Get the # of cores, should be either 1 or 4.
        Num_Cores = psutil.cpu_count(logical=True)

        if (Num_Cores != 1) and (Num_Cores != 4):
            sense.show_message("Unsupported number of cores (not 1 or 4)")
            sleep(5)
            sense.clear()
            exit()

        #   At this point it's OK to run. If using Ping Hosts start its thread so it can ping most,
        #       if not all, of the hosts before the other data is displayed.

        #   If using It's Alive set it up and then check for Ping Hosts.
        if Its_Alive_On:
            Its_Alive_Row_Number = 7            #   It's Alive starts on the top row.
            Its_Alive_Going_Down = True         #   The direction the row is moving.
            Its_Alive_First_Time = True         #   Don't display on first pass, it looks tacky.
                                                #   If using Ping Hosts start the Ping Hosts daemon's thread.
            if Ping_Hosts_Display_On:
                Ping_Hosts_Row = Its_Alive_Row  #   Initialize the row before the asynchronous process starts
                Ping_Hosts_Daemon_Thread = threading.Thread(target=Ping_Hosts_Daemon)
                Ping_Hosts_Daemon_Thread.start()

        #   Display the TV Effect - TV turning on.
        TV_On_Off_Effect(True)

        #   Display the Raspberry Pi logo.
        sense.set_pixels(RPi_Logo_LED_Array)
        sleep(Seconds_to_Display_RPi_Logo)
        sense.clear()

        #   Display the Apple logo.
        sense.set_pixels(Apple_Logo_LED_Array)
        sleep(Seconds_to_Display_Apple_Logo)
        sense.clear()

        #   Set the Low Light to the user's preference.
        sense.low_light = Low_Light_Mode_On

        #   Set the rotation based on the user's preference.
        #   Since the LED array is treated as columns instead of rows
        #       it needs to be rotated -90 (270) degrees.
        sense.set_rotation((LED_Array_Rotation + 270) % 360)

        #   Create the LED array, initialize to all Black.
        LED_Array = Black_LED_Array

        #   Do calculations that only need to be done once.
        CPU_Temp_Divisions = (CPU_Temp_Cutoff_High - CPU_Temp_Cutoff_Low) / 8
        HAT_Temp_Divisions = (HAT_Temp_Cutoff_High - HAT_Temp_Cutoff_Low) / 8


        #   Each column's array offset in LED_Array.
        CPU_Load_Array_Offset = CPU_Load_Display_Column_1 * 8
        Memory_Array_Offset = Memory_Display_Column * 8
        CPU_Temp_Array_Offset = CPU_Temp_Display_Column * 8
        HAT_Temp_Array_Offset = HAT_Temp_Display_Column * 8
        Humidity_Array_Offset = Humidity_Display_Column * 8

        # Loop until ctrl-c is pressed (or program exited in GUI shell).
        while True:

            #   CPU Load:
            if CPU_Load_Display_On:
                CPU_Load_Array_Offset = CPU_Load_Display_Column_1 * 8   #   The column's array offset in LED_Array.
                if CPU_by_Core:
                                                                    #   Get a list of the percent load on each core (CPU).
                    CPU_Loads_List = psutil.cpu_percent(interval=None, percpu=True)
                                                                    #   CPU loads can vary from 0 to 100.0%.
                                                                    #   Since there are 8 LEDs each one is a 12.5% (100/8) increase.
                    #   Loop through the cores.
                    for This_Core in range (0, Num_Cores):
                        This_Core_0_to_8 = int(round((CPU_Loads_List[This_Core] / 12.5),0))
                        if This_Core_0_to_8 == 0:
                            This_Core_Column = Column_For_0_Value
                        else:
                            This_Core_Column = Color_Spectrum_Column[0:This_Core_0_to_8] + Black_Column[0:8-This_Core_0_to_8]
                        if (Num_Cores == 1) and CPU_by_Core_Single_Core_2_Columns:
                                                                     #   For a single core using two columns.
                            LED_Array = LED_Array[0:CPU_Load_Array_Offset:] + This_Core_Column + This_Core_Column + LED_Array[(CPU_Load_Array_Offset + 16):64]
                        else:
                            LED_Array = LED_Array[0:(CPU_Load_Array_Offset + (This_Core * 8))] + This_Core_Column + \
                                        LED_Array[(CPU_Load_Array_Offset + ((This_Core + 1) * 8)):64]
                else:
                                                                    #   Get the combined load across all cores.
                    CPU_Load = psutil.cpu_percent(interval=None, percpu=False)
                    CPU_Load_0_to_8 = int(round((CPU_Load / 12.5),0))
                    if CPU_Load_0_to_8 == 0:
                        New_Load_Column = Column_For_0_Value
                    else:
                        New_Load_Column = Color_Spectrum_Column[0:CPU_Load_0_to_8] + Black_Column[0:8-CPU_Load_0_to_8]
                                                                    #   This one is trickier than the others.
                                                                    #   First grab anything that could be to the left of the CPU columns.
                                                                    #   Then grab the right three columns of the CPU (effectively shifting them left).
                                                                    #   Then add the new column and finally grab anything to the right.
                    LED_Array = LED_Array[0:CPU_Load_Array_Offset] + LED_Array[(CPU_Load_Array_Offset + 8):(CPU_Load_Array_Offset + 32)] + \
                                New_Load_Column + LED_Array[(CPU_Load_Array_Offset + 40):64]

            #   Memory:
            if Memory_Display_On:
                                                                    #   Since this isn't a spectrum it's best placed to separate the % loads from the temperatures.
                                                                    #   Get the memory stats.
                Mem_Stats =  psutil.virtual_memory()                #   The units don't matter since we're only showing the ratio.
                Mem_Available_0_to_8 = int(round ((8 * (Mem_Stats.available / Mem_Stats.total)), 0))
                LED_Array = LED_Array[0:Memory_Array_Offset] + Mem_Avail_Column[0:Mem_Available_0_to_8] + Mem_Unavailable_Column[0:(8-Mem_Available_0_to_8)] + \
                            LED_Array[(Memory_Array_Offset + 8):64]

            #   CPU Temperature:
            if CPU_Temp_Display_On:
                f = open("/sys/class/thermal/thermal_zone0/temp")
                CPU_Temp_1000s_Celsius = f.read()                   #   A string of digits in 1/1000 degrees Celsius, e.g. 45678
                f.close()
                CPU_Temp = int(CPU_Temp_1000s_Celsius) / 1000.0     #   Normalize to decimal Celsius, e.g. 45.678
                if CPU_Temp_Fahrenheit:
                    CPU_Temp = (CPU_Temp * 1.8) + 32                #   Convert Celsius to Fahrenheit.
                                                                    #   Unlike the CPU load, we can't know the exact range of possible temperatures.
                                                                    #   So we need to handle temperatures out of the range of the low..high cutoffs.
                CPU_Temp_0_to_8 = Normalize_to_0_Thru_8(int(round((CPU_Temp - CPU_Temp_Cutoff_Low) / CPU_Temp_Divisions, 0)))
                if CPU_Temp_0_to_8 == 0:
                    CPU_Temp_Column = Column_For_0_Value
                else:
                    CPU_Temp_Column = Color_Spectrum_Column[0:CPU_Temp_0_to_8] + Black_Column[0:(8 - CPU_Temp_0_to_8)]
                LED_Array = LED_Array[0:CPU_Temp_Array_Offset] + CPU_Temp_Column + LED_Array[(CPU_Temp_Array_Offset + 8):64]

            #   HAT Temperature:
            if HAT_Temp_Display_On:
                HAT_Temp = sense.get_temperature()
                if HAT_Temp_Fahrenheit:
                    HAT_Temp = (HAT_Temp * 1.8) + 32                #   Convert Celsius to Fahrenheit.
                                                                    #   Unlike the CPU load, we can't know the exact range of possible temperatures.
                                                                    #   So we need to handle temperatures out of the range of the low..high cutoffs.
                HAT_Temp_0_to_8 = Normalize_to_0_Thru_8(int(round((HAT_Temp - HAT_Temp_Cutoff_Low) / HAT_Temp_Divisions, 0)))
                if HAT_Temp_0_to_8 == 0:
                    HAT_Temp_Column = Column_For_0_Value
                else:
                    HAT_Temp_Column = Color_Spectrum_Column[0:HAT_Temp_0_to_8] + Black_Column[0:(8-HAT_Temp_0_to_8)]
                LED_Array = LED_Array[0:HAT_Temp_Array_Offset] + HAT_Temp_Column +  LED_Array[(HAT_Temp_Array_Offset + 8):64]

            #   Humidity:
            if Humidity_Display_On:
                Humidity_0_to_8 = int(round((sense.get_humidity() / 12.5), 0))
                if Humidity_0_to_8 == 0:                            #   Is this even possible on this planet with this sensor?
                    Humidity_Column = Column_For_0_Value
                else:
                     Humidity_Column = Humidity_LED_Column[0:Humidity_0_to_8] + Black_Column[0:(8 - Humidity_0_to_8)]
                LED_Array = LED_Array[0:Humidity_Array_Offset] + Humidity_Column + LED_Array[(Humidity_Array_Offset + 8):64]


            #   It's Alive (flashing rows)
            if Its_Alive_On:
                if Its_Alive_First_Time:                            #   Ignore the first pass through the loop.
                    Its_Alive_First_Time = False
                else:
                    Its_Alive_LED_Array = LED_Array                 #   Get a copy of the LED array.
                    for This_Pixel in range (0, 8):                 #   Populate each column's pixel through the row.
                        if Ping_Hosts_Display_On:
                            This_Pixel_Color = Ping_Hosts_Row[This_Pixel]
                        else:
                            This_Pixel_Color = Its_Alive_Row[This_Pixel]
                        Pixel_Index = Its_Alive_Row_Number + (This_Pixel * 8)
                        Its_Alive_LED_Array = Its_Alive_LED_Array[0:Pixel_Index] + This_Pixel_Color + Its_Alive_LED_Array[Pixel_Index + 1:64]
                    sense.set_pixels(Its_Alive_LED_Array)           #   Output the modified LED array.
                    sleep(Its_Alive_Seconds_to_Flash)               #   Wait a short time. No need to clear since normal LED array output right after this.
                                                                    #   Use the boolean being 0 or 1 to calculate increment or decrement without an if test.
                    Its_Alive_Row_Number = Its_Alive_Row_Number + 1 - (Its_Alive_Going_Down * 2)
                    if (Its_Alive_Row_Number % 7 ) == 0:            #   Equivalent to: (Its_Alive_Row_Number == 0) or (Its_Alive_Row_Number == 7).
                        Its_Alive_Going_Down = not Its_Alive_Going_Down #   We're at the top or bottom, change direction.

            #   Output the array to the Sense HAT's LEDs.
            sense.set_pixels(LED_Array)
            #   Wait before looping.
            sleep(Seconds_Between_Updates)

    except KeyboardInterrupt:
        #   Either ctrl-c was typed or the GUI shell interrupted execution.
        #   Turn Low Light off while logos are displayed.
        sense.low_light = False
        #   Set the rotation based on the user's preference.
        sense.set_rotation(LED_Array_Rotation)
        #   Display the Apple logo.
        sense.set_pixels(Apple_Logo_LED_Array)
        sleep(Seconds_to_Display_Apple_Logo)
        #   Display the Raspberry Pi logo.
        sense.set_pixels(RPi_Logo_LED_Array)
        sleep(Seconds_to_Display_RPi_Logo)
        #   Display the TV Effect - TV turning off.
        TV_On_Off_Effect(False)
        #   Clear the Sense Hat's display.
        sense.clear()

#############################################################################
#                                                                           #
#   END - Program code.                                                     #
#                                                                           #
#############################################################################
